from collections import defaultdict

import structlog
from structlog.stdlib import LoggerFactory

structlog.configure(logger_factory=LoggerFactory())
LOGGER = structlog.get_logger(__name__)


class Heatmap(object):
    def __init__(self, file_intersection, file_count_map, commit_count):
        """
        Create a Heatmap object.

        :param file_intersection: Map of how files intersect.
        :param file_count_map: Map of how many times files where seen.
        :param commit_count: Number of commits seen.
        """
        self._file_intersection = file_intersection
        self._file_count_map = file_count_map
        self.commit_count = commit_count
        self._normalized_map = None

    @classmethod
    def create_heatmap(cls, repo, test_re, source_re, look_until=None, last_commit=None):
        """
        Create a heatmap for the given repository.

        :param repo: Repository to analyze.
        :param test_re: Regular expression to match tests.
        :param source_re: Regular expression to match source.
        :param look_until: How far back to analyze commits.
        :param last_commit: Process results until command is seen.
        :return: Number of commits visited, Heatmap of repo, count file.
        """
        file_intersection = defaultdict(lambda: defaultdict(int))
        file_count = defaultdict(int)

        LOGGER.debug('searching until', ts=look_until, commit=last_commit)
        commit_count = 0
        for commit in repo.walk_commits(repo.head()):
            LOGGER.debug('Investigating commit', summary=commit.summary(), ts=commit.commit_time,
                         id=commit.id)

            if look_until and commit.commit_time.timestamp() < look_until.timestamp():
                break

            if last_commit and commit.id == last_commit:
                break

            commit_count += 1

            tests_changed = set()
            src_changed = set()
            for path in commit.new_or_changed_files(commit.parent):
                LOGGER.debug('found change', path=path)

                if test_re.match(path):
                    tests_changed.add(path)
                elif source_re.match(path):
                    src_changed.add(path)

            for src in src_changed:
                file_count[src] += 1
                for t in tests_changed:
                    file_intersection[src][t] += 1

        return Heatmap(file_intersection, file_count, commit_count)

    def get_heatmap(self, normalize=False):
        """
        Get a dictionary of the heatmap.

        :param normalize: Return a normalized version.
        :return: Dictionary of the heatmap.
        """
        if normalize:
            return self.normalize()
        return self._file_intersection

    @staticmethod
    def _normalize_row(test_map, src_file_count):
        """
        Normalize the given row.

        :param test_map: Map to normalize.
        :param src_file_count: Number of time the source file has been seen.
        :return: Normalized version of row.
        """
        return {k: v / src_file_count for k, v in test_map.items()}

    def normalize(self):
        """
        Get a normalized version of the heatmap.

        :return: Normalized version of heatmap.
        """
        if not self._normalized_map:
            self._normalized_map = {
                k: self._normalize_row(v, self._file_count_map[k])
                for k, v in self._file_intersection.items()}
        return self._normalized_map



from datetime import datetime, timedelta
import logging
import json
import re

import click

from gittoolbox.git.git_repo import GitRepo
from gittoolbox.heatmap.heatmap import Heatmap


def display_row(heatmap):
    """
    Display the information for one source file.

    :param heatmap: test files mapping
    """
    print(", ".join([f"{k}: {heatmap[k]}" for k in sorted(heatmap, key=heatmap.get, reverse=True)]))


@click.command('git-heatmap')
@click.option('--repo-location', required=True, help='Repository to analyze.')
@click.option('-t', '--test-regex', required=True, help='Regex to match test files.')
@click.option('-s', '--source-regex', required=True, help='Regex to match source files.')
@click.option('--months-back', default=None, type=int, help='Number of months to analyze.')
@click.option('--json', 'write_json', is_flag=True, help='Write output in json.')
@click.option('--normalize', is_flag=True, help='Normalize how often test files are seen.')
@click.option('--verbose', is_flag=True, help='Enable verbose logging.')
@click.option('--to-commit', default=None, help='Compute to the given commit.')
def create(repo_location, test_regex, source_regex, months_back, write_json, normalize, verbose,
           to_commit):
    """
    Create a heat map of which test files have been changed in the same commit as source files.

    This can then be used to help determine which tests would be good to run when a source
    file changes.
    \f

    :param repo_location: Location of git repository.
    :param test_regex: Regex to match test files.
    :param source_regex: Regex to match source files.
    :param months_back: How far back to look.
    :param write_json: Should output be written in json.
    """
    log_level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(level=log_level)

    look_until = None
    if months_back:
        look_until = datetime.now() - timedelta(weeks=(4 * months_back))
    repo = GitRepo.local_repo(repo_location)
    test_re = re.compile(test_regex)
    source_re = re.compile(source_regex)

    heatmap = Heatmap.create_heatmap(repo, test_re, source_re, look_until, to_commit)
    if write_json:
        print(json.dumps(heatmap.get_heatmap(normalize), indent=4))
    else:
        heatmap_dict = heatmap.get_heatmap(normalize)
        for src in heatmap_dict:
            print(f'{src}: ', end='')
            display_row(heatmap_dict[src])

        print(f"Looked at {heatmap.commit_count} commits.")



from typing import Set

from git import Repo, Commit, DiffIndex

import structlog
from structlog.stdlib import LoggerFactory


structlog.configure(logger_factory=LoggerFactory())
LOGGER = structlog.get_logger(__name__)


class GitRepo(object):
    """A git repository."""
    def __init__(self, repo: Repo):
        """
        Create a new GitRepo.
        :param repo: Repository representing repo.
        """
        self._repo = repo

    @classmethod
    def local_repo(cls, path):
        """
        Create a GitRepo for a local repository.
        :param path: Path to local repository.
        :return: GitRepo for local repository.
        """
        return cls(Repo(path))

    def head(self):
        return self._repo.head.commit

    def walk_commits(self, start_commit):
        """
        Walk the commits in the repository.

        :param start_commit: Commit to start walking.
        :return: Generator that walks commits.
        """
        for commit in self._repo.iter_commits(start_commit):
            yield GitCommit(commit)


class GitCommit(object):
    """A git commit object."""
    def __init__(self, commit: Commit):
        """
        Create an object representing a commit.

        :param commit: commit object.
        """
        self._commit = commit

    def summary(self):
        """
        Get a summary of the commit message.
        :return: Summary of commit message.
        """
        return self._commit.message.splitlines()[0]

    @property
    def commit_time(self):
        """
        Get the time the commit was created.
        :return: commit time.
        """
        return self._commit.committed_datetime

    @property
    def id(self):
        return self._commit.hexsha

    @property
    def parent(self):
        """
        Get the parent for this commit.

        Returns the first commit if this is a merge commit.
        :return: Parent of commit.
        """
        LOGGER.debug('getting parents', parents=self._commit.parents)
        return self._commit.parents[0]

    def diff_to_parent(self):
        """
        Compare this commit to its parent.

        :return:
        """
        return GitDiff(self._commit.diff(self.parent))

    def new_or_changed_files(self, commit: Commit) -> Set:
        """
        Get a set of files that were new or changed compared to the given commit.

        :param commit: Commit to compare against.
        :return: Set of files that are new or changed between commits.
        """
        diff = GitDiff(self._commit.diff(commit))
        return {change.b_path for change in diff.new_file_iter()}


class GitDiff(object):
    """A Git diff object."""
    def __init__(self, diff: DiffIndex):
        """
        Create an object representing a diff.
        :param diff: diff object.
        """
        self._diff = diff

    def new_file_iter(self):
        """
        Iterate over file for added changes in the diff.

        :return: Iterator for added files.
        """
        for patch in self._diff.iter_change_type('M'):
            yield patch

        for patch in self._diff.iter_change_type('A'):
            yield patch

        for patch in self._diff.iter_change_type('R'):
            yield patch
